/*
 */

#define __SND_OSS_COMPAT__
#include "../../include/driver.h"
#include "../../include/emu10k1.h"

#define VOL_RANGE_MAX	31

static void update_rear(emu10k1_t *emu, ac97_t *ac97)
{
	int reg = FXGPREGBASE + 0x10;

	if (! emu->rear_active
#if 0
	    || (ac97->regs[AC97_MASTER] & 0x8000)
#endif
	    ) {
		snd_emu10k1_ptr_write(emu, reg, 0, 0);
		snd_emu10k1_ptr_write(emu, reg + 1, 0, 0);
	} else {
		unsigned int val;
#if 0
		val = (ac97->regs[AC97_MASTER] >> 8) & ac97->max_master;
		val = ac97->max_master - val;
		val *= emu->vol_rear[0];
		val *= 0x7fffffff / (VOL_RANGE_MAX * ac97->max_master);
		snd_emu10k1_ptr_write(emu, reg, 0, val);
		val = ac97->regs[AC97_MASTER] & ac97->max_master;
		val = ac97->max_master - val;
		val *= emu->vol_rear[1];
		val *= 0x7fffffff / (VOL_RANGE_MAX * ac97->max_master);
		snd_emu10k1_ptr_write(emu, reg + 1, 0, val);
#else
		val = emu->vol_rear[0] * (0x7fffffff / VOL_RANGE_MAX);
		snd_emu10k1_ptr_write(emu, reg, 0, val);
		val = emu->vol_rear[1] * (0x7fffffff / VOL_RANGE_MAX);
		snd_emu10k1_ptr_write(emu, reg + 1, 0, val);
#endif
	}
}
	
static int snd_emu10k1_volume_rear(snd_kmixer_element_t *element, int w_flag, int *volume)
{
	emu10k1_t *emu;
	ac97_t *ac97;
	int change = 0;
	unsigned long flags;

	emu = snd_magic_cast(emu10k1_t, element->private_data, -ENXIO);
	ac97 = emu->ac97;
	snd_debug_check(ac97 == NULL, -ENXIO);
	
	spin_lock_irqsave(&ac97->reg_lock, flags);
	if (w_flag) {
		change = (volume[0] != emu->vol_rear[0] ||
			  volume[1] != emu->vol_rear[1]);
		emu->vol_rear[0] = volume[0];
		emu->vol_rear[1] = volume[1];
		update_rear(emu, ac97);
	} else {
		volume[0] = emu->vol_rear[0];
		volume[1] = emu->vol_rear[1];
	}
	spin_unlock_irqrestore(&ac97->reg_lock, flags);
	return change;
}

static int snd_emu10k1_mute_rear(snd_kmixer_element_t *element, int w_flag, int *value)
{
	emu10k1_t *emu;
	ac97_t *ac97;
	int change = 0;
	unsigned long flags;

	emu = snd_magic_cast(emu10k1_t, element->private_data, -ENXIO);
	ac97 = emu->ac97;
	snd_debug_check(ac97 == NULL, -ENXIO);

	spin_lock_irqsave(&ac97->reg_lock, flags);
	if (w_flag) {
		change = (*value != emu->rear_active);
		emu->rear_active = *value;
		update_rear(emu, ac97);
	} else {
		*value = emu->rear_active;
	}
	spin_unlock_irqrestore(&ac97->reg_lock, flags);
	return change;
}

static int snd_emu10k1_mixer_group_fx(snd_kmixer_group_t *group,
				      snd_kmixer_file_t *file, int w_flag,
				      snd_mixer_group_t *ugroup)
{
	emu10k1_t *emu;
	int voices[2], sw;
	int change = 0;

	emu = snd_magic_cast(emu10k1_t, group->private_data, -EINVAL);

	if (!w_flag) {
		ugroup->channels = SND_MIXER_CHN_FRONT_LEFT|SND_MIXER_CHN_FRONT_RIGHT;
		ugroup->channels |= SND_MIXER_CHN_REAR_LEFT|SND_MIXER_CHN_REAR_RIGHT;
		ugroup->caps = SND_MIXER_GRPCAP_VOLUME|SND_MIXER_GRPCAP_MUTE;
		ugroup->mute = 0;
		/*
		ugroup->volume.names.front_left = emu->fx_mix[FX_FRONT].vol[0];
		ugroup->volume.names.front_right = emu->fx_mix[FX_FRONT].vol[1];
		*/
		if (! emu->rear_active) {
			ugroup->mute |= (1 << SND_MIXER_CHN_REAR_LEFT);
			ugroup->mute |= (1 << SND_MIXER_CHN_REAR_RIGHT);
		}
		ugroup->volume.names.rear_left = emu->vol_rear[0];
		ugroup->volume.names.rear_right = emu->vol_rear[1];
		ugroup->min = 0;
		ugroup->max = VOL_RANGE_MAX;
	} else {
		/*
		sw =  (ugroup->mute & (1 << SND_MIXER_CHN_FRONT_LEFT)) ? 1 : 0;
		if (snd_emu10k1_mute_front(emu->me_sw_front, 1, &sw)) change = 1;
		voices[0] = ugroup->volume.names.front_left;
		voices[1] = ugroup->volume.names.front_right;
		if (snd_emu10k1_volume_front(emu->me_sw_front, 1, voices)) change = 1;
		*/
		sw =  (ugroup->mute & (1 << SND_MIXER_CHN_REAR_LEFT)) ? 1 : 0;
		if (snd_emu10k1_mute_rear(emu->me_sw_rear, 1, &sw)) change = 1;
		voices[0] = ugroup->volume.names.rear_left;
		voices[1] = ugroup->volume.names.rear_right;
		if (snd_emu10k1_volume_rear(emu->me_sw_rear, 1, voices)) change = 1;
	}
	return change;
}

static void snd_emu10k1_mixer_free_ac97(void *private_data)
{
	emu10k1_t *emu = snd_magic_cast(emu10k1_t, private_data, );
	emu->mixer = NULL;
}

int snd_emu10k1_mixer(emu10k1_t *emu, int device, snd_pcm_t *pcm, snd_kmixer_t **rmixer)
{
	ac97_t ac97;
	snd_kmixer_t *mixer;
	snd_kmixer_group_t *group;
	static struct snd_mixer_element_volume1_range table13_range[2] = {
		{0, VOL_RANGE_MAX, -3450, 1200},
		{0, VOL_RANGE_MAX, -3450, 1200}
	};
	int err, dev;

	*rmixer = NULL;
	memset(&ac97, 0, sizeof(ac97));
	ac97.write = snd_emu10k1_ac97_write;
	ac97.read = snd_emu10k1_ac97_read;
	ac97.private_data = emu;
	ac97.private_free = snd_emu10k1_mixer_free_ac97;
	dev = 0;
	if ((err = snd_ac97_mixer(emu->card, device, &ac97, 1, &dev /* &pcm->device */, &mixer)) < 0)
		return err;
	emu->ac97 = snd_magic_cast(ac97_t, mixer->private_data, -ENXIO);

	emu->rear_active = 0;
	emu->vol_rear[0] = 0;
	emu->vol_rear[1] = 0;

	/* build rear speaker control */
	if ((group = snd_mixer_lib_group_ctrl(mixer, "Rear", 0, SND_MIXER_OSS_UNKNOWN, snd_emu10k1_mixer_group_fx, emu)) == NULL)
		goto __error;
	if ((emu->me_vol_rear = snd_mixer_lib_volume1(mixer, "Rear", 0, 2,
						      table13_range, snd_emu10k1_volume_rear, emu)) == NULL)
		goto __error;
	if ((emu->me_sw_rear = snd_mixer_lib_sw2(mixer, "Rear Switch", 0, snd_emu10k1_mute_rear, emu)) == NULL)
		goto __error;
	if (snd_mixer_group_element_add(mixer, group, emu->me_vol_rear) < 0)
		goto __error;
	if (snd_mixer_group_element_add(mixer, group, emu->me_sw_rear) < 0)
		goto __error;
	if (snd_mixer_element_route_add(mixer, emu->me_vol_rear, emu->me_sw_rear))
		goto __error;
	if (snd_mixer_element_route_add(mixer, emu->me_sw_rear, emu->ac97->me_accu))
		goto __error;

	*rmixer = emu->mixer = mixer;
	return 0;

__error:
	snd_device_free(emu->card, mixer);
	return -ENOMEM;
}
